import importlib as IM
import logging

from .dataobject import (BaseDataObject, DataObject, RegistryEntry,
                         PythonClassDescription, Module, PythonModule, ClassDescription,
                         ClassResolutionFailed, ModuleResolutionFailed)
from .utils import FCN
from .configure import Configurable


# TODO: Move this into mapper or a new mapper_common module
CLASS_REGISTRY_CONTEXT_KEY = 'class_registry_context_id'
'''
.. confval:: class_registry_context_id

    Configuration file key for the URI of the class registry RDF graph context.

    The class registry context holds the mappings between RDF types and Python classes for
    a project or bundle.
'''

CLASS_REGISTRY_CONTEXT_LIST_KEY = 'class_registry_context_list'
'''
.. confval:: class_registry_context_list

    Configuration file key for the list of class registry contexts

    If it is specified, then :confval:`class_registry_context_id` should be searched first
    for class registry entries. The class registry list may be built automatically or not
    defined at all depending on who makes the `Configuration`, but if it is specified with
    this property, then it should be respected.
'''

__all__ = ["Mapper",
           "UnmappedClassException"]

L = logging.getLogger(__name__)


class UnmappedClassException(Exception):
    pass


class ClassRedefinitionAttempt(Exception):
    '''
    Thrown when a `.Mapper.add_class` is called on a class when a class with the same name
    has already been added to the mapper
    '''
    def __init__(self, mapper, maybe_cls, cls):
        super(ClassRedefinitionAttempt, self).__init__(
                'Attempted to add class %s to %s when %s had already been added' % (
                    maybe_cls, mapper, cls))


class Mapper(Configurable):
    '''
    Keeps track of relationships between Python classes and RDF classes

    The mapping this object manages may also be written to the RDF graph as `class
    registry entries <RegistryEntry>`. The entries are written to the "class registry
    context", which can be specified when the Mapper is created.
    '''
    def __init__(self, name=None, class_registry_context=None,
            class_registry_context_list=None, **kwargs):
        '''
        Parameters
        ----------
        name : str, optional
            Name of the mapper for diagnostic/debugging purposes
        class_registry_context : `owmeta_core.context.Context` or str, optional
            The context where mappings should be saved and/or retrieved from. Either the
            context object itself or the ID for it. If not provided, then the class
            registry context ID is looked up from the Mapper's configuration at
            `CLASS_REGISTRY_CONTEXT_KEY`
        class_registry_context_list : list of `owmeta_core.context.Context` or str, optional
            List of contexts or context IDs where registry entries should be retrieved
            from if the class_registry_context doesn't yield a mapping
        **kwargs
            passed to super-classes
        '''
        super(Mapper, self).__init__(**kwargs)

        # Maps full class names (i.e., including the module name) to classes
        self._mapped_classes = dict()

        # Maps RDF types to properties of the related class
        self._rdf_type_table = dict()

        # Modules that have already been loaded
        self.modules = dict()

        if name is None:
            name = hex(id(self))
        self.name = name

        self.__class_registry_context_id = None
        self.__class_registry_context = None
        if isinstance(class_registry_context, str):
            self.__class_registry_context_id = class_registry_context
        else:
            self.__class_registry_context = class_registry_context

        self.__class_registry_context_id_list = None
        self.__class_registry_context_list = []
        if class_registry_context_list:
            if isinstance(class_registry_context_list[0], str):
                self.__class_registry_context_id_list = class_registry_context_list
            else:
                self.__class_registry_context_list = class_registry_context_list

        self._bootstrap_mappings()

    @property
    def class_registry_context(self):
        ''' Context where class registry entries are stored '''
        if self.__class_registry_context is None:
            from .context import Context
            crctx_id = (self.__class_registry_context_id
                    or self.conf.get(CLASS_REGISTRY_CONTEXT_KEY, None))
            if crctx_id is None:
                return None
            crctx = Context(crctx_id, conf=self.conf, mapper=self)
            self.__class_registry_context = crctx
        return self.__class_registry_context

    @property
    def class_registry_context_list(self):
        '''
        Context where class registry entries are retrieved from if
        `class_registry_context` doesn't contain an appropriate entry
        '''
        if self.__class_registry_context_list is None:
            from .context import Context
            crctx_ids = (self.__class_registry_context_id_list
                    or self.conf.get(CLASS_REGISTRY_CONTEXT_LIST_KEY, None))
            if crctx_ids is None:
                return []
            crctxs = []
            for crctx_id in crctx_ids:
                crctxs.append(Context(crctx_id, conf=self.conf, mapper=self).stored)
            self.__class_registry_context_list = crctxs
        return (([self.class_registry_context.stored]
                if self.class_registry_context else []) +
                self.__class_registry_context_list)

    def _bootstrap_mappings(self):
        # Add classes needed for resolving other classes...
        # XXX: Smells off...probably don't want to have to do this.
        self.process_classes(BaseDataObject, DataObject, PythonClassDescription, Module,
                             ClassDescription, PythonModule, RegistryEntry)

    def add_class(self, cls):
        '''
        Add a class to the mapper

        Parameters
        ----------
        cls : type
            The class to add to the mapper

        Raises
        ------
        ClassRedefinitionAttempt
            Thrown when `add_class` is called on a class when a class with the same name
            has already been added to the mapper
        '''
        cname = FCN(cls)
        maybe_cls = self._lookup_class(cname)
        if maybe_cls is not None:
            if maybe_cls is cls:
                return False
            else:
                raise ClassRedefinitionAttempt(self, maybe_cls, cls)
        L.debug("Adding class %s@0x%x", cls, id(cls))

        self._mapped_classes[cname] = cls
        L.debug('parents %s', parents_str(cls))

        if hasattr(cls, 'on_mapper_add_class'):
            cls.on_mapper_add_class(self)

        # This part happens after the on_mapper_add_class has run since the
        # class has an opportunity to set its RDF type based on what we provide
        # in the Mapper.
        self._rdf_type_table[cls.rdf_type] = cls
        return True

    def load_module(self, module_name):
        """ Loads the module. """
        module = self.lookup_module(module_name)
        if not module:
            module = IM.import_module(module_name)
            return self.process_module(module_name, module)
        else:
            return module

    def process_module(self, module_name, module):
        self.modules[module_name] = module
        self._module_load_helper(module)
        return module

    def process_class(self, *classes):
        for c in classes:
            self.add_class(c)

    process_classes = process_class

    def lookup_module(self, module_name):
        return self.modules.get(module_name, None)

    def _check_is_good_class_registry(self, cls):
        module = IM.import_module(cls.__module__)
        if hasattr(module, cls.__name__):
            return

        ymc = getattr(module, '__yarom_mapped_classes__', None)
        if ymc and cls in ymc:
            return

        L.warning(('While saving the registry entry of {}, we found that its'
                  ' module, {}, does not have "{}" in its'
                  ' namespace').format(cls, cls.__module__, cls.__name__))

    def save(self):
        crctx = self.class_registry_context
        if crctx is None:
            raise Exception(f'{self}.class_registry_context is unset.'
                    ' Cannot save class registry entries')
        self.declare_python_class_registry_entry(*self._rdf_type_table.values())
        crctx.save()
        crctx.save_imports()

    def declare_python_class_registry_entry(self, *classes):
        crctx = self.class_registry_context
        if crctx is None:
            raise Exception(f'{self}.class_registry_context is unset.'
                    ' Cannot declare class registry entries')
        for cls in classes:
            crctx(cls).declare_class_registry_entry()

    def load_registry_entries(self):
        crctx = self.class_registry_context.stored
        return crctx(RegistryEntry)().load()

    def resolve_class(self, uri, context):
        '''
        Look up the Python class for the given URI recovered from the given `~.Context`

        Parameters
        ----------
        uri : rdflib.term.URIRef
            The URI to look up
        context : .Context
            The context the URI was found in. May affect which Python class is returned.
        '''

        # look up the class in the registryCache
        c = self._rdf_type_table.get(uri)
        if c is not None:
            return c
        # otherwise, attempt to load into the cache by
        # reading the RDF graph.

        if self.class_registry_context is None:
            L.warning('%s.class_registry_context is unset.'
                    ' Cannot resolve class for "%s"',
                    self, uri)
            return None
        resolved_class = None
        for crctx in self.class_registry_context_list:
            resolved_class = self._resolve_class(uri, crctx)
            if resolved_class:
                break

        if resolved_class:
            self.add_class(resolved_class)
        return resolved_class

    def _resolve_class(self, uri, crctx):
        re = crctx(RegistryEntry)()
        re.rdf_class(uri)
        cd = crctx(PythonClassDescription)()
        re.class_description(cd)
        c = None

        for cd_l in cd.load():
            try:
                c = cd_l.resolve_class()
            except ClassResolutionFailed as e:
                if isinstance(e.__cause__, ModuleResolutionFailed):
                    L.warn('_resolve_class: Did not find module', exc_info=True)
                    continue
            if c is not None:
                break

            # Fall-back class resolution
            class_name = cd_l.name()
            if class_name is None:
                L.warning('_resolve_class: Could not find a class name attached to'
                          ' %s', cd_l)
                continue
            moddo = cd_l.module()
            if moddo is None:
                # Try loading the module using the more generic ClassDescription:module
                # relationship instead. It's acceptable as long as the type is
                # PythonModule
                L.warning('_resolve_class: Could not find a module attached via'
                          ' PythonClassDescription:module to %s for the class named %s.'
                          ' Trying ClassDescription:module instead...', cd_l, class_name)
                moddo = ClassDescription.module(cd_l)()
                if not isinstance(moddo, PythonModule):
                    L.warning('_resolve_class: Could not find a module attached to'
                              ' %s for the class named %s', cd_l, class_name)
                    continue
            mod = moddo.resolve_module()
            L.warning('_resolve_class: Did not find class %s in %s', class_name, mod.__name__)
            ymc = getattr(mod, '__yarom_mapped_classes__', None)
            if not ymc:
                L.warning('_resolve_class: No __yarom_mapped_classes__ in %s, so cannot look up %s',
                        mod.__name__, class_name)
                continue

            matching_classes = tuple(mc for mc in ymc
                                     if mc.__name__ == class_name)
            if not matching_classes:
                L.warning('_resolve_class: Did not find class %s in %s.__yarom_mapped_classes__',
                        class_name, mod.__name__)
                continue

            c = matching_classes[0]
            if len(matching_classes) > 1:
                L.warning('_resolve_class: More than one class has the same name in'
                        ' __yarom_mapped_classes__ for %s, so we are picking'
                        ' the first one as the resolved class among %s',
                        mod, matching_classes)
            break
        return c

    def _module_load_helper(self, module):
        # TODO: Make this class selector pluggable
        return self.handle_mapped_classes(getattr(module, '__yarom_mapped_classes__', ()))

    def handle_mapped_classes(self, classes):
        res = []
        for cls in classes:
            if isinstance(cls, type) and self.add_class(cls):
                res.append(cls)
        return res

    def lookup_class(self, cname):
        """ Gets the class corresponding to a fully-qualified class name """
        ret = self._lookup_class(cname)
        if ret is None:
            raise UnmappedClassException((cname,))
        return ret

    def _lookup_class(self, cname):
        return self._mapped_classes.get(cname, None)

    def mapped_classes(self):
        for c in self._mapped_classes.values():
            yield c

    def __str__(self):
        if self.name is not None:
            return f'{type(self).__name__}(name="{str(self.name)}")'
        else:
            return super(Mapper, self).__str__()


class _ClassOrderable(object):
    def __init__(self, cls):
        self.cls = cls

    def __eq__(self, other):
        self.cls is other.cls

    def __gt__(self, other):
        res = False
        ocls = other.cls
        scls = self.cls
        if issubclass(ocls, scls) and not issubclass(scls, ocls):
            res = True
        elif issubclass(scls, ocls) == issubclass(ocls, scls):
            res = scls.__name__ > ocls.__name__
        return res

    def __lt__(self, other):
        res = False
        ocls = other.cls
        scls = self.cls
        if issubclass(scls, ocls) and not issubclass(ocls, scls):
            res = True
        elif issubclass(scls, ocls) == issubclass(ocls, scls):
            res = scls.__name__ < ocls.__name__
        return res


def parents_str(cls):
    return ", ".join(p.__name__ + '@' + hex(id(p)) for p in cls.mro())



from six import with_metaclass
from .data import DataUser
from .contextualize import Contextualizable, ContextualizableClass


class CustomPropertyMeta(ContextualizableClass):
    def __init__(self, name, bases, dct):
        super(CustomPropertyMeta, self).__init__(name, bases, dct)
        self.__context = None

    @property
    def context(self):
        return self.__context

    @context.setter
    def context(self, ctx):
        self.__context = ctx


class CustomProperty(with_metaclass(CustomPropertyMeta, Contextualizable, DataUser)):
    """ Store a value associated with a DataObject

    Properties can be be accessed like methods. A method call like::

        a.P()

    for a property ``P`` will return values appropriate to that property for
    ``a``, the `owner` of the property.

    Parameters
    ----------
    owner : owmeta_core.dataobject.DataObject
        The owner of this property
    name : string
        The name of this property. Can be accessed as an attribute like::

            owner.name
    """

    # Indicates whether the CustomProperty is multivalued
    multiple = False

    def __init__(self, name=False, owner=False, **kwargs):
        super(CustomProperty, self).__init__(**kwargs)
        self.owner = owner
        self.linkName = name
        if self.owner:
            if name:
                setattr(self.owner, name, self)

    @property
    def values(self):
        return []

    def get(self, *args):
        """ Get the things which are on the other side of this property

        The return value must be iterable. For a ``get`` that just returns
        a single value, an easy way to make an iterable is to wrap the
        value in a tuple like ``(value,)``.

        Derived classes must override.
        """
        # This should run a query or return a cached value
        raise NotImplementedError()

    def set(self, *args, **kwargs):
        """ Set the value of this property

        Derived classes must override.
        """
        # This should set some values and call DataObject.save()
        raise NotImplementedError()

    get_terms = get

    def one(self):
        """
        Returns a single value for the `CustomProperty` whether or not it is multivalued.
        """

        r = self.get()
        return next(iter(r), None)

    def has_value(self):
        """
        Returns true if the `CustomProperty` has any values set on it.

        This may be defined differently for each property
        """
        return True

    def __call__(self, *args, **kwargs):
        """
        If arguments are passed to the `CustomProperty`, its :meth:`set` method is called.
        Otherwise, the `get` method is called. If the ``multiple`` member for the
        `CustomProperty` is set to `True`, then a Python :py:func:`set` containing the
        associated values is returned. Otherwise, a single value is returned.
        """

        if len(args) > 0 or len(kwargs) > 0:
            self.set(*args, **kwargs)
            return self
        else:
            r = self.get(*args, **kwargs)
            if self.multiple:
                return set(r)
            else:
                return next(iter(r), None)



from __future__ import print_function
import sys
import types
import argparse
import copy as _copy
import functools
from textwrap import wrap
import logging

from .utils import FCN
from .docscrape import parse as npdoc_parse
from .command_util import IVar, SubCommand


# TODO: Use `inspect` module for getting argument names so we aren't depending on
# docstrings
from .cli_common import (INSTANCE_ATTRIBUTE,
                         METHOD_NAMED_ARG,
                         METHOD_NARGS,
                         METHOD_KWARGS)

from .cli_hints import CLI_HINTS


L = logging.getLogger(__name__)

ARGUMENT_TYPES = {
    'int': int
}
''' Map from parameter types to type constructors for parsing arguments '''


_KVLIST_ARG = '_KVLIST_ARG'


class CLIUserError(Exception):
    '''
    An error which the user would have to correct.

    Typically caused by invalid user input
    '''


def _method_runner(runner, key):
    L.debug('Getting method %s from %s', key, runner)
    method = getattr(runner, key)

    @functools.wraps(method)
    def _f(*args, **kwargs):
        return method(*args, **kwargs)
    return _f


def _sc_runner(sub_mapper, sub_runner):
    def _f():
        return sub_mapper.apply(sub_runner)
    return _f


class CLIArgMapper(object):
    '''
    Stores mappings for arguments and maps them back to the part of the object
    they come from
    '''
    def __init__(self):
        self.mappings = dict()
        self.methodname = None
        self.runners = dict()
        ''' Mapping from subcommand names to functions which run for them '''

        self.named_arg_count = dict()

        self.argparser = None
        # A special little mapper just for callable runners
        self.runner_mapper = None

    def apply(self, runner):
        '''
        Applies the collected arguments to the runner by calling methods and traversing
        the object attributes as required

        Parameters
        ----------
        runner : object
            Target of the command and source of argument and method names

        See Also
        --------
        CLICommandWrapper : accepts a runner argument in its ``__init__`` method
        '''
        iattrs = self.get(INSTANCE_ATTRIBUTE)
        kvpairs = self.get(METHOD_KWARGS)
        kvs = list(kv.split('=', 1) for kv in next(iter(kvpairs.values()), ()))
        kwargs = {k: v for k, v in kvs}

        args = self.get_list(METHOD_NAMED_ARG)
        if not args:
            kwargs.update(self.get(METHOD_NAMED_ARG))

        kvlist = self.get(_KVLIST_ARG)
        for key, vals in kvlist.items():
            kwargs[key] = list(kv.split('=', 1) for kv in vals)

        try:
            # There is, at most, one nargs entry.
            nargs = next(iter(self.get(METHOD_NARGS).values()))
        except StopIteration:
            nargs = ()

        L.debug('Looking up runner %s from %s', self.methodname, self.runners)
        runmethod = self.runners.get(self.methodname, None)

        def continuation():
            argcount = self.named_arg_count.get(self.methodname)
            if nargs and args and argcount is not None and len(args) != argcount:
                # This means we have passed in positional arguments, and we have a
                # variable-length option, but we have not filled out all of the arguments
                # necessary to cleanly apply the runmethod since Python would think we're
                # trying to apply some arguments twice.
                #
                # We *could* support a slightly richer set of options here, but it's probably
                # not worth it...
                #
                # Also, this is a programmer error. End-users shouldn't hit this
                raise Exception('Missing arguments to method ' + str(self.methodname))

            return runmethod(*(tuple(args) + tuple(nargs)), **kwargs)

        for k, v in iattrs.items():
            setattr(runner, k, v)

        if callable(runner) and self.runner_mapper:
            if runmethod is not None:
                runner._next = continuation
            return self.runner_mapper.apply(runner)

        if runmethod is None:
            self.argparser.print_help(file=sys.stderr)
            print(file=sys.stderr)
            raise CLIUserError('Please specify a sub-command')

        return continuation()

    def get(self, key):
        return {k[1]: self.mappings[k] for k in self.mappings if k[0] == key and
                self.mappings[k] is not None}

    def get_list(self, key):
        keys = sorted((k for k in self.mappings.keys() if k[0] == key), key=lambda it: it[2])
        last = -1
        for k in keys:
            if k[2] - last != 1:
                return []
            last = k[2]
        return [self.mappings[k] for k in keys]

    def __str__(self):
        return type(self).__name__ + '(' + str(self.mappings) + ')'


class CLIStoreAction(argparse.Action):
    ''' Interacts with the CLIArgMapper '''

    def __init__(self, mapper, key, index=-1, mapped_name=None, *args, **kwargs):
        '''
        Parameters
        ----------
        mapper : CLIArgMapper
            CLI argument to Python mapper
        key : str
            Indicates what kind of argument is being mapped. One of `.INSTANCE_ATTRIBUTE`,
            `.METHOD_NAMED_ARG`, `.METHOD_KWARGS`, `.METHOD_NARGS`
        index : int
            Argument index. Used for maintaining the order of arguments when passed to the
            runner
        mapped_name : str
            The name to map to. optional.
        *args
            passed to `~argparse.Action`
        **kwargs
            passed to `~argparse.Action`
        '''
        super(CLIStoreAction, self).__init__(*args, **kwargs)
        if self.nargs == 0:
            raise ValueError('nargs for store actions must be > 0; if you '
                             'have nothing to store, actions such as store '
                             'true or store const may be more appropriate')
        if self.const is not None and self.nargs != argparse.OPTIONAL:
            raise ValueError('nargs must be %r to supply const' % argparse.OPTIONAL)

        self.mapper = mapper
        self.key = key
        self.name = mapped_name or self.dest
        self.index = index

    def __call__(self, parser, namespace, values, option_string=None):
        self.mapper.mappings[(self.key, self.name, self.index)] = values
        setattr(namespace, self.dest, values)


class CLIStoreTrueAction(CLIStoreAction):
    '''
    Action for storing `True` when a given option is provided
    '''
    def __init__(self, *args, **kwargs):
        '''
        Parameters
        ----------
        *args
            passed to `~.CLIStoreAction`
        **kwargs
            passed to `~.CLIStoreAction`
        '''
        super(CLIStoreTrueAction, self).__init__(*args, **kwargs)
        self.nargs = 0

    def __call__(self, parser, namespace, values, option_string=None):
        super(CLIStoreTrueAction, self).__call__(parser, namespace, True, option_string)


class CLIAppendAction(CLIStoreAction):
    '''
    Extends CLIStoreAction to append to a set of accumulated values

    Used for recording a `dict`
    '''
    def __call__(self, parser, namespace, values, option_string=None):
        '''
        Parameters
        ----------
        parser
            Ignored
        namespace : argparse.Namespace
            Namespace to add to
        values : str
            Value to add
        '''
        items = _copy.copy(_ensure_value(namespace, self.dest, []))
        if isinstance(values, list):
            # This can happen because of a kwargs-type argument that has nargs='*'
            items += values
        else:
            items.append(values)
        self.mapper.mappings[(self.key, self.name, -1)] = items
        setattr(namespace, self.dest, items)


class CLISubCommandAction(argparse._SubParsersAction):
    '''
    Action for sub-commands

    Extends the normal action for sub-parsers to record the subparser name in a mapper
    '''

    def __init__(self, mapper, *args, **kwargs):
        '''
        Parameters
        ----------
        mapper : CLIArgMapper
            CLI argument to Python mapper
        *args
            Passed on to `argparse._SubParsersAction`
        **kwargs
            Passed on to `argparse._SubParsersAction`
        '''
        super(CLISubCommandAction, self).__init__(*args, **kwargs)
        self.mapper = mapper

    def __call__(self, *args, **kwargs):
        if self.mapper.methodname is not None:
            raise ValueError('More than one sub command has been specified!'
                             'Attempted to set {} when {} had already been'
                             ' set.'.format(self.dest, self.mapper.methodname))

        self.mapper.methodname = args[2][0]
        super(CLISubCommandAction, self).__call__(*args, **kwargs)


NOT_SET = object()


def _ensure_value(namespace, name, value):
    if getattr(namespace, name, None) is None:
        setattr(namespace, name, value)
    return getattr(namespace, name)


class CLICommandWrapper(object):
    '''
    Wraps an object such that it can be used in a command line interface
    '''

    def __init__(self, runner, mapper=None, hints=None, hints_map=None, program_name=None):
        '''
        Parameters
        ----------
        runner : object
            An object that provides the methods to be invoked
        mapper : CLIArgMapper
            Stores the arguments and associated runners for the command. A mapper is
            created if none is provided. optional
        hints : dict
            A multi-level dict describing how certain command line arguments get turned
            into attributes and method arguments. If `hints` is not provided, the hints
            are looked up by the runner's fully-qualified class name in `hints_map`. optional
        hints_map : dict
            A multi-level dict describing how certain command line arguments get turned
            into attributes and method arguments. Defaults to `CLI_HINTS <.cli_hints>`. optional
        program_name : str
            The name of the top-level program. Uses `sys.argv[0] <sys.argv>` if not provided.
            optional
        '''
        self.runner = runner
        self.mapper = CLIArgMapper() if mapper is None else mapper
        self.hints_map = hints_map or CLI_HINTS
        self.hints = self.hints_map.get(FCN(type(runner)), {}) if hints is None else hints
        self.program_name = program_name

    def extract_args(self, val):
        '''
        Extract arguments from the method or class docstring

        In the return value (see below), the ``summary`` is a `str` used in listing out
        sub-commands.  The ``detail`` is for the sub-command usage information and should,
        generally, include the ``summary``. The ``params`` are a list
        `~owmeta_core.docscrape.ParamInfo` objects describing the parameters.

        Parameters
        ----------
        val : object
            The object with the documentation

        Returns
        -------
        tuple
            a triple, ``(summary, detail, params)``
        '''
        docstring = getattr(val, '__doc__', '')
        if not docstring:
            docstring = ''
        npdoc = npdoc_parse(docstring)
        params = npdoc.get('parameters')
        paragraphs = self._split_paras(npdoc.get('desc') or '')

        if len(paragraphs) >= 1:
            summary = paragraphs[0]
        else:
            summary = ''

        detail = '\n \n'.join('\n'.join(wrap(x, width=80)) for x in paragraphs if x)

        return summary, detail, params

    def _split_paras(self, docstring):
        paragraphs = []
        temp = ''
        for ln in docstring.split('\n'):
            ln = ln.strip()
            if ln:
                temp += '\n' + ln
            else:
                if temp:
                    paragraphs.append(temp.strip())
                temp = ''
        if temp:
            paragraphs.append(temp.strip())

        return paragraphs

    def parser(self, parser=None):
        '''
        Generates the argument parser's arguments

        Parameters
        ----------
        parser : argparse.ArgumentParser
            The parser to add the arguments to. optional: will create a parser if none is
            given
        '''
        if parser is None:
            doc = getattr(self.runner, '__doc__', None)
            if doc:
                cmd_summary, _, _ = self.extract_args(self.runner)
            else:
                cmd_summary = None
            parser = argparse.ArgumentParser(prog=self.program_name, description=cmd_summary)
        self.mapper.argparser = parser
        for key, val in vars(self.runner).items():
            if not key.startswith('_') and key not in self.hints.get('IGNORE', ()):
                parser.add_argument('--' + key, help=key.__doc__)

        _sp = [None]

        def sp():
            if _sp[0] is None:
                _sp[0] = parser.add_subparsers(dest='subparser', mapper=self.mapper,
                                               action=CLISubCommandAction)
            return _sp[0]

        runner_type_attrs = dict()
        runner_type = type(self.runner)
        for x in dir(runner_type):
            if x.startswith('_') or x in self.hints.get('IGNORE', ()):
                continue
            runner_type_attrs[x] = getattr(runner_type, x)
        if '__call__' in dir(runner_type):
            # Handle sub-commands which are, themselves, callable. Summary and details
            # must be specified on the sub-command class docstring and would have already
            # been handled by the ``isinstance(val, SubCommand)`` case below
            _, _, params = self.extract_args(runner_type.__call__)
            saved_mapper = self.mapper
            self.mapper = CLIArgMapper()
            self._handle_method(self.program_name, parser, '__call__',
                    runner_type.__call__, params)
            saved_mapper.runner_mapper = self.mapper
            self.mapper = saved_mapper
            # reset our mapper so the methods and such
        for key, val in sorted(runner_type_attrs.items()):
            if isinstance(val, (types.FunctionType, types.MethodType)):
                command_name = key.replace('_', '-')
                summary, detail, params = self.extract_args(val)
                subparser = sp().add_parser(command_name,
                        help=summary,
                        description=detail,
                        formatter_class=argparse.RawDescriptionHelpFormatter)
                self._handle_method(command_name, subparser, key, val, params)
            elif isinstance(val, property):
                doc = getattr(val, '__doc__', None)
                parser.add_argument('--' + key, help=doc,
                                    action=CLIStoreAction,
                                    key=INSTANCE_ATTRIBUTE,
                                    mapper=self.mapper)
            elif isinstance(val, SubCommand):
                summary, detail, params = self.extract_args(val)
                sub_runner = getattr(self.runner, key)
                sub_mapper = CLIArgMapper()

                command_name = key.replace('_', '-')
                self.mapper.runners[command_name] = _sc_runner(sub_mapper, sub_runner)

                subparser = sp().add_parser(command_name, help=summary, description=detail)
                type(self)(sub_runner, sub_mapper, hints_map=self.hints_map).parser(subparser)
            elif isinstance(val, IVar):
                doc = getattr(val, '__doc__', None)
                var_hints = self.hints.get(key) if self.hints else None
                if val.default_value:
                    if doc:
                        doc += '. Default is ' + repr(val.default_value)
                    else:
                        doc = 'Default is ' + repr(val.default_value)
                # NOTE: we have a default value from the val, but we don't
                # set it here -- IVars return the defaults ... by default
                arg_kwargs = dict(help=doc,
                                  action=CLIStoreAction,
                                  key=INSTANCE_ATTRIBUTE,
                                  mapper=self.mapper)
                if val.value_type == bool:
                    arg_kwargs['action'] = CLIStoreTrueAction
                names = None if var_hints is None else var_hints.get('names')
                if names is None:
                    names = ['--' + key.replace('_', '-')]
                parser.add_argument(*names, **arg_kwargs)

        return parser

    def _handle_method(self, command_name, subparser, key, val, params):
        sc_hints = self.hints.get(key) if self.hints else None
        meth = getattr(self.runner, key)
        positional_arg_count = meth.__code__.co_argcount
        self.mapper.runners[command_name] = _method_runner(self.runner, key)
        argcount = 0
        for pindex, param in enumerate(params):
            action = CLIStoreAction
            if param.val_type == 'bool':
                action = CLIStoreTrueAction

            atype = ARGUMENT_TYPES.get(param.val_type)

            arg = param.name
            arg_cli_name = arg.replace('_', '-')
            desc = param.desc
            if arg.startswith('**'):
                arg_hints = self._arg_hints(sc_hints, METHOD_KWARGS, arg[2:])
                names = None if arg_hints is None else arg_hints.get('names')
                if names is None:
                    names = ['--' + arg_cli_name[2:]]
                argument_args = dict(action=CLIAppendAction,
                                     mapper=self.mapper,
                                     key=METHOD_KWARGS,
                                     type=atype,
                                     help=desc)
                if arg_hints:
                    nargs = arg_hints.get('nargs')
                    if nargs is not None:
                        argument_args['nargs'] = nargs
                subparser.add_argument(*names,
                                       **argument_args)
            elif arg.startswith('*'):
                subparser.add_argument(arg_cli_name[1:],
                                       action=action,
                                       nargs='*',
                                       key=METHOD_NARGS,
                                       mapper=self.mapper,
                                       type=atype,
                                       help=desc)
            else:
                arg_hints = self._arg_hints(sc_hints, METHOD_NAMED_ARG, arg)
                names = None if arg_hints is None else arg_hints.get('names')
                if names is None:
                    names = ['--' + arg_cli_name]
                index = pindex
                if positional_arg_count >= pindex:
                    index = -1
                argument_args = dict(action=action,
                                     key=METHOD_NAMED_ARG,
                                     mapper=self.mapper,
                                     index=index,
                                     mapped_name=arg,
                                     type=atype,
                                     help=desc)
                if arg_hints:
                    nargs = arg_hints.get('nargs')
                    if nargs is not None:
                        argument_args['nargs'] = nargs

                    is_kvlist = arg_hints.get('kvlist')
                    if is_kvlist:
                        argument_args['key'] = _KVLIST_ARG

                    action_hint = arg_hints.get('action')
                    hinted_action = None
                    if 'append' == action_hint:
                        hinted_action = CLIAppendAction
                    elif 'store_true' == action_hint:
                        hinted_action = CLIStoreTrueAction
                    elif 'store' == action_hint:
                        hinted_action = CLIStoreAction

                    if hinted_action is not None:
                        argument_args['action'] = hinted_action

                subparser.add_argument(*names,
                                       **argument_args)
                argcount += 1
        self.mapper.named_arg_count[key] = argcount

    def _arg_hints(self, sc_hints, atype, key):
        return None if sc_hints is None else sc_hints.get((atype, key))

    def main(self, args=None, argument_callback=None, argument_namespace_callback=None):
        '''
        Runs in a manner suitable for being the 'main' method for a command
        line interface: parses arguments (as would be done with the result of
        `parser`) from sys.argv or the provided args list and executes the
        commands specified therein

        Parameters
        ----------
        args : list
            the argument list to parse. optional
        argument_callback : callable
            a callback to add additional arguments to the command line. optional
        argument_namespace_callback : callable
            a callback to handle the parsed arguments to the command line. optional
        '''

        parser = self.parser()
        if argument_callback:
            argument_callback(parser)
        ns = parser.parse_args(args=args)
        if argument_namespace_callback:
            argument_namespace_callback(ns)
        return self.mapper.apply(self.runner)



import shutil

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest


def format_table(dat, header=None, pref_widths=None, default_termwidth=400):
    dat = list(dat)
    if not dat:
        return ''

    ncols = len(dat[0])

    widths = None
    avg_widths = None
    if header:
        if ncols != len(header):
            raise Exception('Width of header, {}, does not'
                            ' match width of data, {}'.format(len(header), ncols))

        widths = tuple(_max_width(d) for d in header)
        avg_widths = tuple(widths)

    for ridx, row in enumerate(dat):
        if len(row) != ncols:
            raise Exception('Row widths are not consistent.'
                            ' Expected {}, but found row width of {}'.format(ncols, len(row)))
        these_widths = tuple(_max_width(d) for d in row)
        if widths is None:
            widths = tuple(these_widths)
            avg_widths = tuple(these_widths)
        else:
            widths = tuple(max(v) for v in zip(widths, these_widths))
            avg_widths = tuple((a * (ridx + 1) + c)/(ridx + 2) for a, c in zip(avg_widths, these_widths))

    if pref_widths:
        if ncols != len(pref_widths):
            raise Exception('Width of adjustment preferences,'
                            ' {},  does not match width of data, {}'.format(len(pref_widths), ncols))

        pref_widths = list(pref_widths)
    else:
        max_col_width = max(widths)
        pref_widths = list((max_col_width - x) / max_col_width for x in avg_widths)

    try:
        termwidth, _ = shutil.get_terminal_size((default_termwidth, 0))
    except AttributeError:
        termwidth = default_termwidth

    prefsorted_widths = sorted(zip(pref_widths, range(ncols)), key=lambda x: x[0])
    new_widths = list(widths)
    while sum(new_widths) + ncols - 1 >= termwidth:
        selection = max(prefsorted_widths, key=lambda x: new_widths[x[1]] * x[0])
        new_widths[selection[1]] -= 1

    widths = new_widths

    fmt = ' '.join('{:' + str(w) + '}' for w in widths)

    lines = []
    for row in dat:
        row_strings = list(zip_longest(*(format(f).split('\n') for f, wdth in zip(row, widths)),
                                   fillvalue=''))
        for m in row_strings:
            lines.append(fmt.format(*(it[:wdth] for it, wdth in zip(m, widths))))

    if header:
        w = max(len(m) for m in lines)
        lines.insert(0, '-' * w)
        lines.insert(0, fmt.format(*(format(f)[:wdth] for f, wdth in zip(header, widths))))

    return '\n'.join(lines)


def _max_width(s):
    return max(len(line) for line in str(s).split('\n'))



class RDFTypeResolver:
    '''
    Handles mapping between RDF graphs and Python types
    '''

    def __init__(self, default_type, type_resolver, id2object_translator, deserializer):
        """
        Parameters
        ----------
        default_type : :class:`str`, :class:`rdflib.term.URIRef`
            If no type is retrieved from the graph, this will be the type selected
        type_resolver : callable : (rdflib.graph.Graph, [rdflib.term.URIRef], rdflib.term.URIRef or None) -> rdflib.term.URIRef
            This callable (e.g., function) receives a graph, all the types found for an
            identifier, and the "base" type sought, which constrains the result to be a
            sub-type of the base, and returns a single identifier for a type that
            `id2object_translator` can translate into an object
        id2object_translator : callable : (rdflib.term.URIRef, rdflib.term.URIRef, owmeta_core.context.Context) -> object
            This callable (e.g., function) receives an identifier for an object
            and an identifier for the object's type and returns an object
            corresponding to the identifier and type
        deserializer : callable : (:class:`rdflib.term.Literal`) -> object
            This callable (e.g., function) receives a literal and turns it
            into an object
        """
        self.base_type = default_type
        self.type_resolver = type_resolver
        self.id2ob = id2object_translator
        self.deserializer = deserializer



import abc
import types
from weakref import WeakValueDictionary

import wrapt


class AbstractBaseContextualizable(abc.ABC):
    '''
    Abstract base class for contextualizables

    Any class with an attribute `contextualize` with a Function value is recognized as a
    subclass
    '''
    @classmethod
    def __subclasshook__(cls, subclassP):
        contextualize = getattr(subclassP, 'contextualize', None)
        if contextualize is not None:
            return isinstance(contextualize, types.FunctionType) or isinstance(contextualize, types.MethodType)
        else:
            return NotImplemented


class BaseContextualizable(object):
    '''
    Helper base-class for contextualizable objects. Caches contextualized objects returned
    from `contextualize_augment`
    '''
    def __init__(self, *args, **kwargs):
        super(BaseContextualizable, self).__init__(*args, **kwargs)

        if not hasattr(self, '_contexts'):
            self._contexts = WeakValueDictionary()

    @property
    def context(self):
        return None

    def contextualize(self, context):
        '''
        Return an object with the given context. If the provided ``context`` is
        `None`, then `self` MUST be returned unmodified. Prefer to override
        `contextualize_agument` which will be called from this method.

        It is generally not correct to set a field on the object and return the
        same object as this would change the context for other users of the
        object. Also, returning a copy of the object is usually inappropriate
        for mutable objects. Immutable objects may maintain a 'context'
        property and return a copy of themselves with that property set to the
        provided ``context`` argument.
        '''
        ctxd = self._contexts.get(context)
        if ctxd is not None:
            return ctxd
        ctxd = self.contextualize_augment(context)
        self._contexts[context] = ctxd
        return ctxd

    def decontextualize(self):
        '''
        Return the object with all contexts removed. Sub-classes should override.
        '''
        return self

    def add_contextualization(self, context, contextualization):
        '''
        Manually add a contextualized object to the cache

        Parameters
        ----------
        context : Context
            The context of the object
        contextualization : object
            The contextualized version of the object
        '''
        try:
            self._contexts[context] = contextualization
        except AttributeError:
            self._contexts = WeakValueDictionary()
            self._contexts[context] = contextualization

    def contextualize_augment(self, context):
        '''
        For sub-classes to override: Return an object with the given context. If the
        provided ``context`` is `None`, then `self` MUST be returned unmodified.

        Returns
        -------
        object
            the contextualized object
        '''
        return self


AbstractBaseContextualizable.register(BaseContextualizable)


class Contextualizable(BaseContextualizable):
    '''
    A `BaseContextualizable` with the addition of a default behavior of setting
    the context from the class's 'context' attribute. This generally requires
    that for the metaclass of the Contextualizable that a 'context' data
    property is defined. For example::

        >>> class AMeta(ContextualizableClass):
        ...     @property
        ...     def context(self):
        ...         return self.__context
        ...
        ...     @context.setter
        ...     def context(self, ctx):
        ...         self.__context = ctx

        >>> class A(six.with_metaclass(Contextualizable)):
        ...     pass

    '''

    def __new__(cls, *args, **kwargs):
        #This is defined so that the __init__ method gets a contextualized
        #instance, allowing for statements made in __init__ to be contextualized.

        ores = super(Contextualizable, cls).__new__(cls)
        # XXX: This shouldn't really ever be the property...
        if not isinstance(cls.context, property):
            ores.context = cls.context
            ores._contexts = WeakValueDictionary()
            ores.add_contextualization(cls.context, ores)
            res = ores
        else:
            ores.context = None
            res = ores

        return res

    @property
    def context(self):
        return self.__context

    @context.setter
    def context(self, ctx):
        if isinstance(ctx, property):
            raise Exception('An attempt was made to set a property as the context. This is likely unintended')
        self.__context = ctx


UNSET = object()


def contextualize_metaclass(context, self):
    class _H(type(self)):
        def __init__(self, name, bases, dct):
            super(_H, self).__init__(name, bases, dct)
            self.__ctx = context

        def __repr__(self):
            return self.__name__ + '.contextualize_class(' + repr(context) + ')'

        @property
        def context(self):
            return self.__ctx
    # Setting the name just for debugging...don't care much about the other
    # attributes for now.
    _H.__name__ = 'Ctxd_Meta_' + self.__name__

    return _H


def get_wrapped(self):
    return super(ContextualizingProxy, self).__getattribute__('__wrapped__')


class ContextualizingProxy(wrapt.ObjectProxy):
    '''
    Extension of `.ObjectProxy` such that the "self" argument refers to the proxy rather
    than the wrapped object in methods. The `context` attribute refers to the context on
    the proxy rather than on the wrapped object.
    '''
    __slots__ = ('_self_context', '_self_overrides')

    def __init__(self, ctx, *args, **kwargs):
        '''
        Parameters
        ----------
        ctx : .Context
            The context for this proxy
        *args
            Passed to `wrapt.ObjectProxy`
        **kwargs
            Passed to `wrapt.ObjectProxy`
        '''
        super(ContextualizingProxy, self).__init__(*args, **kwargs)
        self._self_context = ctx
        self._self_overrides = dict()

    def add_attr_override(self, name, override):
        self._self_overrides[name] = override

    def __getattribute__(self, name):
        # This behavior is what I would expect, but wrapt doesn't work this way...
        # General note: This method should never call itself directly although
        # it may call itself indirectly through a descriptor. Use
        # object.__getattribute__ or super(ContextualizingProxy, self).__getattribute__
        # as appropriate for attribute accesses from within
        if name == 'context':
            return super(ContextualizingProxy, self).__getattribute__('_self_context')

        override = super(ContextualizingProxy, self).__getattribute__('_self_overrides').get(name, UNSET)
        if override is not UNSET:
            return override
        wrapped = None
        if name not in ('__wrapped__', '__factory__', '__class__'):
            k = UNSET
            if name in type(self).__dict__:
                k = type(self).__dict__[name]
            else:
                wrapped = get_wrapped(self)
                for t in type(wrapped).mro():
                    if name in t.__dict__:
                        k = t.__dict__[name]
                        break

            if k is not UNSET:
                if hasattr(k, '__get__'):
                    if not hasattr(k, '__set__'):
                        # We have to check the __wrapped__. Don't check our
                        # self since all we have is a context.
                        try:
                            return k.__get__(self, type(self))
                        except AttributeError:
                            # The __wrapped__ doesn't have the named attribute
                            # Pass in this proxy to the descriptor so that
                            # methods, etc.  can access their context

                            # Classmethods are special-cased. We mostly don't do
                            # anything to the class of a proxied object, and we want
                            # classmethods to 'just work', so for this case, we pass in
                            # the wrapped's type
                            if isinstance(k, classmethod):
                                wrapped = get_wrapped(self) if wrapped is None else wrapped
                                return k.__get__(wrapped, type(wrapped))
                            else:
                                raise
                    # it's a data descriptor
                    elif isinstance(k, classmethod):
                        wrapped = get_wrapped(self) if wrapped is None else wrapped
                        return k.__get__(wrapped, type(wrapped))
                    else:
                        return k.__get__(self, type(self))
                else:
                    try:
                        wrapped = get_wrapped(self) if wrapped is None else wrapped
                        return object.__getattribute__(wrapped, name)
                    except AttributeError:
                        return k

        return super(ContextualizingProxy, self).__getattribute__(name)

    def __setattr__(self, name, value):
        # This was copied from wrapt/wrappers.py with the addition noted below
        if name.startswith('_self_'):
            object.__setattr__(self, name, value)
        elif name == '__wrapped__':
            raise AttributeError('Cannot set wrapped after initialization')
        elif name == '__qualname__':
            setattr(get_wrapped(self), name, value)
            object.__setattr__(self, name, value)
        else:
            if name in self._self_overrides:
                self._self_overrides[name] = value
                return
            # Added compared to wrapt.
            mro = type(self).mro()
            for x in mro:
                attr = x.__dict__.get(x, None)
                if hasattr(attr, '__set__'):
                    attr.__set__(self, value)
                    break
            else: # no break
                setattr(get_wrapped(self), name, value)

    def __call__(self, *args, **kwargs):
        # Omitted by default and only included in CallableObjectProxy by wrapt.
        # Dunno why.
        return get_wrapped(self).__call__.__func__(self, *args, **kwargs)

    def __repr__(self):
        return 'ContextualizingProxy({}, {})'.format(repr(self._self_context),
                                                     repr(self.__wrapped__))


class ContextualizableClass(type):
    ''' A super-type for contextualizable classes

    Attributes
    ----------
    context_carries : tuple of str
        When defining a specialized contextualizable class, you may want to define some
        attribute on the class that is only set if it's declared directly in the class
        body (e.g., by using `property` and name mangling). However, by default,
        contextualization creates a subclass and you may want your property to be
        "carried" into the new context. You can achieve this by declaring
        `context_carries` with the names of attributes that should be carried through a
        contextualization.
    '''
    context_carries = ()

    def __new__(self, name, typ, dct):
        res = super(ContextualizableClass, self).__new__(self, name, typ, dct)
        res.__contexts = WeakValueDictionary()
        return res

    def __init__(self, name, bases, dct):
        super().__init__(name, bases, dct)
        carries = set(type(self).context_carries)
        for base in type(self).__bases__:
            base_carries = getattr(base, 'context_carries', ())
            carries |= set(base_carries)

        self.context_carries = tuple(carries)

    def __getattribute__(self, name):
        # This method is optimized to save a comparison in the common case
        if name in ('contextualize', 'contextualize_augment'):
            if name == 'contextualize_augment':
                name = 'contextualize_class_augment'
            else:
                name = 'contextualize_class'
        return super(ContextualizableClass, self).__getattribute__(name)

    def contextualize_class(self, context):
        ctxd = self.__contexts.get(context)
        if ctxd is not None:
            return ctxd
        ctxd = self.contextualize_class_augment(context)
        self.__contexts[context] = ctxd
        return ctxd

    def contextualize_class_augment(self, context, **kwargs):
        if context is None:
            return self
        _H = contextualize_metaclass(context, self)

        for cc in self.context_carries:
            if hasattr(self, cc):
                kwargs[cc] = getattr(self, cc)

        res = _H(self.__name__, (self,), dict(class_context=self.definition_context, **kwargs))
        res.__module__ = self.__module__
        return res


AbstractBaseContextualizable.register(ContextualizableClass)


def contextualized_new(ccls):
    def _helper(cls, *args, **kwargs):
        ores = super(ccls, cls).__new__(cls)
        if cls.context is not None:
            res = ores.contextualize(cls.context)
            res.__init__ = type(ores).__init__.__get__(res, type(ores))
            type(ores).__init__(res, *args, **kwargs)
        else:
            res = ores
        return res
    return _helper


class _ContextualzingProxyMetaType(type(ContextualizingProxy)):
    def __new__(self, name, typ, dct, oclasstyp):
        res = super(_ContextualzingProxyMetaType, self).__new__(self, name, typ, dct)
        res._oct = oclasstyp
        res.__module__ = oclasstyp.__module__
        return res

    def __init__(self, name, typ, dct, oclasstyp):
        self._oct = oclasstyp

    def __getattr__(self, name):
        try:
            return super(_ContextualzingProxyMetaType, self).__getattr__(name)
        except AttributeError:
            return getattr(self._oct, name)


def decontextualize_helper(obj):
    '''
    Removes contexts from a ContextualizingProxy
    '''
    ret = obj
    while isinstance(ret, ContextualizingProxy):
        ret = get_wrapped(ret)
    return contextualize_helper(None, ret, True)


def contextualize_helper(context, obj, noneok=False):
    '''
    Does some extra stuff to make access to the type of a ContextualizingProxy
    work more-or-less like access to the the wrapped object
    '''
    if not noneok and context is None:
        return obj

    ctx = getattr(obj, 'context', None)
    if ctx is not None and ctx is context:
        return obj

    # Copy our special properties into the class so that they
    # always take precedence over attributes of the same name added
    # during construction of a derived class. This is to save
    # duplicating the implementation for them in all derived classes.

    pclass_dct = dict()
    for k, v in vars(obj.__class__).items():
        if k not in ('__wrapped__', '__name__', '__doc__',
                     '__module__', '__weakref__', '__dict__',
                     '__init__'):
            if hasattr(v, '__get__'):
                pclass_dct[k] = v
            else:
                pclass_dct[k] = proxy_to_X(obj.__class__, k)

    newtyp = _ContextualzingProxyMetaType('CtxProxyClass_' + obj.__class__.__name__,
                                          (ContextualizingProxy,),
                                          pclass_dct,
                                          type(obj.__class__))
    res = newtyp(context, obj)
    obj._contexts[context] = res
    return res


class proxy_to_X(object):
    __slots__ = ('_oclass', '_key')

    def __init__(self, oclass, key):
        self._oclass = oclass
        self._key = key

    def __get__(self, o, typ):
        if o is None:
            return getattr(self._oclass, self._key)
        else:
            raise AttributeError()

    def __str__(self):
        return 'proxy_to_' + self._key

    def __repr__(self):
        return 'contextualize.proxy_to_X({}, {})'.format(repr(self._oclass), repr(self._key))



# -*- coding: utf-8 -*-
"""
Common utilities for translation, massaging data, etc., that don't fit
elsewhere in owmeta_core
"""
import functools
import importlib
import re

__all__ = ['grouper', 'slice_dict']

PROVIDER_PATH_FORMAT = r'''
(?P<module>(?:\w+)(?:\.\w+)*)
:
(?P<provider>(?:\w+)(?:\.\w+)*)'''

PROVIDER_PATH_RE = re.compile(PROVIDER_PATH_FORMAT, flags=re.VERBOSE)


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    while True:
        l = []
        try:
            for x in args:
                l.append(next(x))
        except Exception:
            pass
        yield l
        if len(l) < n:
            break


def slice_dict(d, s):
    return {k: v for k, v in d.items() if k in s}


def FCN(cls):
    return str(cls.__module__) + '.' + str(cls.__name__)


def aslist(fun):
    @functools.wraps(fun)
    def wrapper(*args, **kwargs):
        return list(fun(*args, **kwargs))
    return wrapper


UNSET = object()


def getattrs(obj, names, default=UNSET):
    p = obj
    try:
        for x in names:
            p = getattr(p, x)
        return p
    except AttributeError:
        if default is UNSET:
            raise
        return default


def retrieve_provider(provider_path):
    '''
    Look up a "provider" specified by a string.

    Path to an object that provides something. The format is similar to that for
    setuptools entry points: ``path.to.module:path.to.provider.callable``. Notably,
    there's no name and "extras" are not supported.

    Parameters
    ----------
    provider_path : str
        The path to the provider

    Returns
    -------
    object
        The provider

    Raises
    ------
    ValueError
        The `provider_path` format doesn't match the expected pattern
    AttributeError
        Some element in the path is missing
    '''
    md = PROVIDER_PATH_RE.match(provider_path)
    if not md:
        raise ValueError('Format of the provider path is incorrect')
    module = md.group('module')
    provider = md.group('provider')
    m = importlib.import_module(module)
    attr_chain = provider.split('.')
    return getattrs(m, attr_chain)


def ellipsize(s, max_length):
    t = s[:max_length]
    if t != s:
        if len(t) <= 1:
            return t
        return t[:-1] + '…'
    return t



from itertools import chain
import logging

import rdflib as R
from rdflib.namespace import RDFS, RDF

from .utils import FCN
from .rdf_utils import UP, DOWN, transitive_subjects
from .ranged_objects import InRange


L = logging.getLogger(__name__)


class ZeroOrMore(object):
    def __init__(self, identifier, predicate, index, direction=DOWN):
        self.identifier = identifier
        self.predicate = predicate
        self.direction = direction
        self.direction = direction
        self.index = index

    def __repr__(self):
        return "{}({}, {}, {})".format(FCN(type(self)),
                                       repr(self.identifier),
                                       repr(self.predicate),
                                       repr(self.direction))


class SubClassModifier(ZeroOrMore):

    def __init__(self, rdf_type):
        super().__init__(rdf_type, R.RDFS.subClassOf, 2, UP)

    def __repr__(self):
        return FCN(type(self)) + '(' + repr(self.identifier) + ')'


class SubPropertyOfModifier(ZeroOrMore):

    def __init__(self, rdf_property):
        super().__init__(rdf_property, R.RDFS.subPropertyOf, 1, direction=UP)

    def __repr__(self):
        return FCN(type(self)) + '(' + repr(self.identifier) + ')'


def rdfs_subclassof_zom_creator(target_type):
    '''
    Creates a function used by `ZeroOrMoreTQLayer` to determine if a query needs to be
    augmented to retrieve sub-classes of a *given* RDF type
    '''
    def helper(triple):
        if target_type == triple[2] is not None and triple[1] == R.RDF.type:
            return SubClassModifier(triple[2])
    return helper


def rdfs_subpropertyof_zom(super_property):
    '''
    Argument to `ZeroOrMoreTQLayer`. Adds sub-properties of the given property to triple
    queries
    '''
    def helper(triple):
        if triple[1] == super_property:
            return SubPropertyOfModifier(super_property)
    return helper


def rdfs_subclassof_zom(triple):
    '''
    Argument to `ZeroOrMoreTQLayer`. Adds sub-classes to triple queries for an rdf:type
    '''
    if triple[2] is not None and triple[1] == R.RDF.type:
        return SubClassModifier(triple[2])


def rdfs_subclassof_subclassof_zom_creator(rdf_type):
    def helper(triple):
        '''
        Argument to `ZeroOrMoreTQLayer`. Adds sub-classes to triple queries for an
        rdfs:subClassOf
        '''
        if (triple[1] == R.RDFS.subClassOf and
                triple[2] is not None and
                (rdf_type == triple[2] or
                    (isinstance(triple[2], list) and rdf_type in triple[2]))):
            return SubClassModifier(triple[2])
    return helper


class TQLayer(object):
    '''
    Triple Query Layer. Wraps a graph or another `TQLayer` to do something to the
    `triples` and `triples_choices` query or the result of the query.
    '''
    _NADA = object()

    def __init__(self, nxt=None):
        '''
        Parameters
        ----------
        nxt : TQLayer or rdflib.graph.Graph
            The "next" or "lower" layer that this layer modifies
        '''
        self.next = nxt

    def triples(self, qt, context=None):
        return self.next.triples(qt)

    def triples_choices(self, qt, context=None):
        return self.next.triples_choices(qt)

    def __contains__(self, qt):
        '''
        This should be overridden -- the default implementation just asks the next layer
        if it contains the triple.
        '''
        return qt in self.next

    def __getattr__(self, attr):
        '''
        By default, if this layer doesn't have the given attribute, then the attribute
        will be looked up on the next layer.
        '''
        res = getattr(super(TQLayer, self), attr, TQLayer._NADA)
        if res is TQLayer._NADA:
            return getattr(self.next, attr)

    def __repr__(self):
        return FCN(type(self)) + '(' + repr(self.next) + ')'

    def __str__(self):
        return FCN(type(self)) + '(' + str(self.next) + ')'


class TerminalTQLayer(object):
    '''
    A TQLayer that has no "next". May be useful to create a layer that stands in place of
    a `~rdflib.graph.Graph`.
    '''

    @property
    def next(self):
        raise AttributeError(str(type(self)) + ' has no next layer')

    @next.setter
    def next(self, val):
        raise AttributeError(str(type(self)) + ' has no next layer')

    def triples(self, qt, context=None):
        raise NotImplementedError()

    def triples_choices(self, qt, context=None):
        raise NotImplementedError()


class RangeTQLayer(TQLayer):
    '''
    A layer that understands ranges in the object position of a triple.

    If the next layer has the `supports_range_queries` attribute set to `True`, then the
    range is passed down as-is
    '''

    def triples(self, query_triple, context=None):
        if isinstance(query_triple[2], InRange):
            in_range = query_triple[2]
            if in_range.defined:
                if getattr(self.next, 'supports_range_queries', False):
                    return self.next.triples(query_triple, context)
                else:
                    qt = (query_triple[0], query_triple[1], None)
                    return set(x for x in self.next.triples(qt, context) if in_range(x[2]))
            else:
                qt = (query_triple[0], query_triple[1], None)
                return self.next.triples(qt, context)
        else:
            return self.next.triples(query_triple, context)

    def triples_choices(self, query_triple, context=None):
        if isinstance(query_triple[2], InRange):
            in_range = query_triple[2]
            qt = (query_triple[0], query_triple[1], None)
            if in_range.defined:
                # XXX: Assuming triples_choices does not also support range
                # queries.
                return set(x for x in self.next.triples_choices(qt, context) if in_range(x[2]))
            else:
                return self.next.triples_choices(qt, context)
        else:
            return self.next.triples_choices(query_triple, context)


class ZeroOrMoreTQLayer(TQLayer):
    def __init__(self, transformer, *args):
        '''
        Parameters
        ----------
        transformer : `callable`
            Takes a triple and returns an object describing the relationship or `None`.
            If an object is returned it must have `predicate`, `identifier`,
            `direction`, and `index` attributes.
            - `identifier` is the identifier to start from
            - `predicate` is the predicate to traverse
            - `direction` is the direction of traversal: Either
              `~owmeta_core.rdf_utils.DOWN` for subject -> object or `~owmeta_core.rdf_utils.UP`
              for object -> subject
            - `index` is the index in the triple for which a closure should be looked up
        *args : other arguments
            Go to `TQLayer` init
        '''
        super(ZeroOrMoreTQLayer, self).__init__(*args)
        self._tf = transformer

    def triples(self, query_triple, context=None):
        match = self._tf(query_triple)
        if not match:
            return self.next.triples(query_triple, context)
        qx = list(query_triple)
        if match.identifier is not None:
            matches = list(transitive_subjects(self.next,
                                             match.identifier,
                                             match.predicate,
                                             context,
                                             match.direction))
            qx[match.index] = matches
        L.debug('augmented query %s', qx)
        results = self.next.triples_choices(tuple(qx), context)
        return self._zom_result_helper(results, match, context, set(matches))

    def triples_choices(self, query_triple, context=None):
        match = self._tf(query_triple)
        if not match:
            L.debug('No match %s', query_triple)
            return self.next.triples_choices(query_triple, context)
        qx = list(query_triple)
        iters = []
        # XXX: We should, maybe, apply some stats or heuristics here to determine which list to iterate over.
        if isinstance(match.identifier, list):
            assert isinstance(qx[match.index], list), ('If there is more than one'
            ' matching identifier, the list in the query triple must be in the same'
            ' position')

            matches = set(qx[match.index])
            for match_id in match.identifier:
                matches |= set(transitive_subjects(self.next,
                                                   match_id,
                                                   match.predicate,
                                                   context,
                                                   match.direction,
                                                   matches))
            qx[match.index] = list(matches)
            iters.append(self.next.triples_choices(tuple(qx), context))
        elif match.identifier is not None:
            matches = set(transitive_subjects(self.next,
                                              match.identifier,
                                              match.predicate,
                                              context,
                                              match.direction))
            for sub in matches:
                qx[match.index] = sub
                iters.append(self.next.triples_choices(tuple(qx), context))
        else:
            matches = None
            iters.append(self.next.triples_choices(query_triple, context))
        return self._zom_result_helper(chain(*iters), match, context, matches)

    def _zom_result_helper(self, results, match, context, limit):
        zomses = dict()
        direction = DOWN if match.direction is UP else DOWN
        predicate = match.predicate
        index = match.index
        L.debug('ZeroOrMoreTQLayer: start %r direction %s', match, direction)
        # The results from the original query are augmented here to "entail" results in
        # the "reverse" direction that are implied by the "forward" direction. For
        # instance, if I request everything with a type that's rdfs:Resource, I'll get all
        # type statements we have subclass relationships for with the modified query, but
        # I'll be missing the inferred types. We rectify that below
        #
        for tr in results:
            L.debug('ZeroOrMoreTQLayer: match %r result %r', match, tr)
            zoms = zomses.get(tr[index])
            if zoms is None:
                zoms = set(transitive_subjects(self.next, tr[index], predicate, context, direction))
                if limit:
                    zoms = zoms & limit
                zomses[tr[index]] = zoms
            for z in zoms:
                yield tuple(x if x is not tr[index] else z for x in tr)

    def __contains__(self, query_triple):
        try:
            next(self.triples(query_triple))
            return True
        except StopIteration:
            return False


class ContainerMembershipIsMemberTQLayer(TQLayer):
    '''
    Adds a triple into the results for rdfs:subPropertyOf(rdfs:member) relationships for all
    known ContainerMembershipProperty instances
    '''
    def triples(self, query_triple, context=None):
        iters = [self.next.triples(query_triple, context)]
        if query_triple[1] == RDFS.subPropertyOf and query_triple[2] == RDFS.member:
            iters.append((t[0], RDFS.subPropertyOf, RDFS.member)
                    for t in self.next.triples((None, RDF.type, RDFS.ContainerMembershipProperty), context))
        return chain(*iters)

    def triples_choices(self, query_triple, context=None):
        iters = [self.next.triples_choices(query_triple, context)]
        if (query_triple[1] == RDFS.subPropertyOf and
                (query_triple[2] == RDFS.member or
                    (isinstance(query_triple[2], list) and RDFS.member in query_triple[2]))):
            iters.append((t[0], RDFS.subPropertyOf, RDFS.member)
                    for t in self.next.triples((None, RDF.type, RDFS.ContainerMembershipProperty), context))
        return chain(*iters)


_default_tq_layers_list = [
    RangeTQLayer,
]


def default_tq_layers(base):
    res = base
    for layer in reversed(_default_tq_layers_list):
        res = layer(res)
    return res



"""
For declaring inverse properties of GraphObjects
"""


InverseProperties = dict()


class InversePropertyMixin(object):
    """
    Mixin for inverse properties.

    Augments Property methods to update inverse properties as well
    """

    def set(self, other):
        ip_key = (self.owner_type, self.linkName)
        ip = InverseProperties.get(ip_key)
        if ip:
            rhs_cls, rhs_linkName = ip.other(*ip_key)
            if isinstance(other, rhs_cls):
                rhs_prop = getattr(other.contextualize(self.context), rhs_linkName)
                super(InversePropertyMixin, rhs_prop).set(self.owner)
        return super(InversePropertyMixin, self).set(other)

    def unset(self, other):
        ip_key = (self.owner_type, self.linkName)
        ip = InverseProperties.get(ip_key)
        if ip:
            rhs_cls, rhs_linkName = ip.other(*ip_key)
            if isinstance(other, rhs_cls):
                rhs_prop = getattr(other, rhs_linkName)
                ctxd_rhs_prop = rhs_prop.contextualize(self.context)
                super(InversePropertyMixin, ctxd_rhs_prop).unset(self.owner)
        return super(InversePropertyMixin, self).unset(other)


class InverseProperty(object):

    def __init__(self, lhs_class, lhs_linkName, rhs_class, rhs_linkName):
        self.lhs_class = lhs_class
        self.rhs_class = rhs_class

        self.lhs_linkName = lhs_linkName
        self.rhs_linkName = rhs_linkName
        InverseProperties[(lhs_class, lhs_linkName)] = self
        InverseProperties[(rhs_class, rhs_linkName)] = self

    def other(self, cls, name):
        if issubclass(cls, self.lhs_class) and self.lhs_linkName == name:
            return (self.rhs_class, self.rhs_linkName)
        elif issubclass(cls, self.rhs_class) and self.rhs_linkName == name:
            return (self.lhs_class, self.lhs_linkName)
        raise InversePropertyException('The property ({}, {}) has no inverse in {}'.format(cls, name, self))

    def __repr__(self):
        return 'InverseProperty({},{},{},{})'.format(self.lhs_class,
                                                     self.lhs_linkName,
                                                     self.rhs_class,
                                                     self.rhs_linkName)


class InversePropertyException(Exception):
    pass



import logging

from .configure import Configurable
from .capability import Capable, provide
from .utils import retrieve_provider


CAPABILITY_PROVIDERS_KEY = 'capability.providers'


class CapableConfigurable(Capable, Configurable):
    '''
    Helper class for `.Capable` objects that are also `.Configurable`

    Takes the providers from the :confval:`capability.providers` configuration value and
    calls `provide` with the resulting providers. If the value is unset or empty, then
    `provide` will not be called.

    .. confval:: capability.providers

         a list of `"provider path" <.utils.retrieve_provider>` strings or `Providers
         <.capability.Provider>`
    '''
    def __init__(self, *args, **kwargs):
        '''
        Raises
        ------
        NoProviderAvailable
            if any of the needed capabilities cannot be provided
        '''
        super().__init__(*args, **kwargs)
        conf_providers = self.conf.get(CAPABILITY_PROVIDERS_KEY, [])
        if conf_providers:
            providers = []
            for p in conf_providers:
                if isinstance(p, str):
                    provider = retrieve_provider(p)()
                else:
                    provider = p
                if not provider:
                    raise Exception('No provider found for ' + p)
                providers.append(provider)
            provide(self, providers)
        else:
            logging.debug("No providers for %s in %r", self, self.conf)



from os.path import exists
import logging

from git import Repo


L = logging.getLogger(__name__)


class GitRepoProvider:
    '''
    Provides a project repository for `~.command.OWM` backed by a Git repository
    '''
    def __init__(self):
        self._repo = None
        self.base = None

    def init(self, base=None):
        base = self.base if not base else base
        self._repo = Repo.init(base)

    def add(self, files):
        self.repo().index.add(files)

    def remove(self, files, recursive=False):
        self.repo().index.remove(files, r=recursive)

    def reset(self, *paths):
        from git.refs.head import HEAD
        repo = self.repo()
        HEAD(repo).reset(paths=paths)
        paths = [p for p in paths if exists(p)]
        repo.index.checkout(paths=paths, force=True)

    def commit(self, msg):
        self.repo().index.commit(msg)

    def repo(self):
        if self._repo is None:
            self._repo = Repo(self.base)
        return self._repo

    def clone(self, url, base, progress=None, **kwargs):
        '''
        Parameters
        ----------
        url : str
            URL to clone from
        base : str
            Directory to clone into
        progress : `tqdm.tqdm`-like
            Must support a `progress.update` method accepting the amount to add to total
            progress (see https://tqdm.github.io/docs/tqdm/#update)
        '''
        # Techincally, url and base can be "path-like", but we don't make it part of the
        # formal interface by documenting that
        if progress is not None:
            try:
                progress = _CloneProgress(progress)
            except TypeError:
                L.warning("Progress reporter does not have the necessary interface for "
                " reporting clone progress", exc_info=True)
        Repo.clone_from(url, base, progress=progress, **kwargs)

    def is_dirty(self, path=None):
        return self.repo().is_dirty(path=path)


class _CloneProgress(object):

    def __init__(self, progress_reporter):
        try:
            updater = progress_reporter.update
        except AttributeError:
            raise TypeError("Progress reporter must have an 'update' method")
        else:
            if not callable(updater):
                raise TypeError("Progress reporter 'update' attribute does not appear to"
                        " be callable")
        self.pr = progress_reporter
        try:
            self.pr.unit = 'objects'
        except AttributeError:
            pass

        self._opcode = 0

    def __call__(self, op_code, cur_count, max_count=None, message=''):
        if op_code != self._opcode:
            self.pr.n = 0
        if max_count is not None:
            self.pr.total = max_count
        self.pr.update(cur_count - self.pr.n)



from __future__ import print_function
import logging
from itertools import chain

import six

from .ranged_objects import InRange
from .rdf_utils import UP, DOWN
from .rdf_query_modifiers import default_tq_layers

L = logging.getLogger(__name__)

__all__ = [
    "GraphObject",
    "GraphObjectQuerier",
    "GraphObjectChecker",
    "ComponentTripler",
    "IdentifierMissingException",
]

EMPTY_SET = frozenset([])


class Variable(int):
    """
    A marker used in `GraphObjectQuerier` for variables in a query
    """


class _Range(InRange):
    pass


class GraphObject(object):
    """
    An object which can be included in the object graph.

    An abstract base class.
    """

    def __init__(self, **kwargs):
        super(GraphObject, self).__init__(**kwargs)
        self.properties = []
        self.owner_properties = []

    @property
    def identifier(self):
        """ Must return an object representing this object or else
        raise an Exception. """
        raise NotImplementedError()

    @property
    def defined(self):
        """ Returns true if an :meth:`identifier` would return an identifier
        """
        raise NotImplementedError()

    def variable(self):
        """ Must return a `~owmeta_core.graph_object.Variable` object that identifies this
        `.GraphObject` in queries.

        The variable can be randomly generated when the object is created and
        stored in the object.
        """
        raise NotImplementedError()

    @property
    def idl(self):
        if self.defined:
            return self.identifier
        else:
            return self.variable()

    def __hash__(self):
        raise NotImplementedError()

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif isinstance(other, GraphObject):
            return self.idl == other.idl

    def __lt__(self, other):
        if isinstance(other, GraphObject):
            return self.idl < other.idl
        else:
            return id(self) < id(other)


class GraphObjectChecker(object):
    '''
    Checks the graph of defined GraphObjects for
    '''

    def __init__(self, query_object, graph, sort_first=False):
        self.query_object = query_object
        self.graph = graph

    def __call__(self):
        tripler = ComponentTripler(self.query_object)
        L.debug('GOC: Checking %s', self.query_object)
        for x in sorted(tripler()):
            if x not in self.graph:
                L.debug('GOC: Failed on %s', x)
                return False
        return True


class GraphObjectValidator(object):
    def __init__(self, query_object, graph):
        self.query_object = query_object
        self.graph = graph

    def __call__(self):
        return True


class GraphObjectQuerier(object):

    """ Performs queries for objects in the given graph.

    The querier queries for objects at the center of a star graph. In SPARQL,
    the query has the form::

        SELECT ?x WHERE {
            ?x  <p1> ?o1 .
            ?o1 <p2> ?o2 .
             ...
            ?on <pn> <a> .

            ?x  <q1> ?n1 .
            ?n1 <q2> ?n2 .
             ...
            ?nn <qn> <b> .
        }

    It is allowed that ``<px> == <py>`` for ``x != y``.

    Queries such as::

        SELECT ?x WHERE {
            ?x  <p1> ?o1 .
             ...
            ?on <pn>  ?y .
        }

    or::

        SELECT ?x WHERE {
            ?x  <p1> ?o1 .
             ...
            ?on <pn>  ?x .
        }

    or::

        SELECT ?x WHERE {
            ?x  ?z ?o .
        }

    or::

        SELECT ?x WHERE {
            ?x  ?z <a> .
        }

    are not supported and will be ignored without error.
    """

    def __init__(self, q, graph, hop_scorer=None):
        """
        Call the GraphObjectQuerier object to perform the query.

        Parameters
        ----------
        q : :class:`GraphObject`
            The object which is queried on
        graph : :class:`object`
            The graph from which the objects are queried. Must implement a
            method :meth:`triples` that takes a triple pattern, ``t``, and
            returns a set of triples matching that pattern. The pattern for
            ``t`` is ``t[i] = None``, 0 <= i <= 2, indicates that the i'th
            position can take any value.

            The ``graph`` method can optionally implement the 'range query' 'interface':
            the graph must have a property ``supports_range_queries`` equal to `True` and
            :meth:`triples` must accept an `~owmeta_core.ranged_objects.InRange` object in
            the object position of the query triple, but only for literals
        hop_scorer : callable
            Returns a score for a hop (a four-tuple, (subject, predicate,
            object, target)) indicating how selective the query would be for
            that hop, with lower numbers being more selective. In general the
            score should only take the given hop into account -- it should not
            take previously given hops into account when calculating a score.
        """

        self.query_object = q
        L.debug('GOQ graph %s', graph)
        self.graph = default_tq_layers(graph)
        self.results = dict()
        self.triples_cache = dict()
        self.hop_scorer = hop_scorer

    def do_query(self):
        L.debug('do_query: Graph %s', self.graph)
        if self.query_object.defined:
            L.debug('do_query: Query object %s is already defined', self.query_object)
            gv = GraphObjectChecker(self.query_object, self.graph)
            if gv():
                return set([self.query_object.identifier])
            else:
                L.debug('do_query: Query graph does not align with the backing graph')
                return EMPTY_SET

        qp = _QueryPreparer(self.query_object)
        paths = qp()
        if len(paths) == 0:
            return EMPTY_SET
        h = self.merge_paths(paths)
        if L.isEnabledFor(logging.DEBUG):
            L.debug('do_query: merge_paths_result:\n%s', self._format_merged(h))
        return self.query_path_resolver(h)

    def merge_paths(self, l):
        """ Combines a list of lists into a multi-level table with
        the elements of the lists as the keys. For given::

            [[a, b, c], [a, b, d], [a, e, d]]

        merge_paths returns::

            {a: {b: {c: {},
                     d: {}},
                 e: {d: {}}}}
        """
        res = dict()
        if L.isEnabledFor(logging.DEBUG):
            L.debug("merge_paths: path %s", _format_paths(l))
        for x in l:
            if len(x) > 0:
                tmp = res.get(x[0], [])
                tmp.append(x[1:])
                res[x[0]] = tmp

        for x in res:
            res[x] = self.merge_paths(res[x])

        return res

    def _format_merged(self, merge, depth=0):
        sio = six.StringIO()
        for triple, remainder in merge.items():
            idx = triple.index(None)
            other_idx = 0 if (idx == 2) else 2
            print((depth * 4 * ' ') + str(triple[1]) + '->' + str(triple[other_idx]), file=sio)
            print(self._format_merged(remainder, depth+1), file=sio, end='')
        return sio.getvalue()

    def query_path_resolver(self, path_table):
        join_args = []
        goal = None
        for hop in sorted(path_table.keys(), key=self.score):
            L.debug("query_path_resolver: hop %s", hop)
            goal = hop[3]
            self._qpr_helper(path_table[hop], hop, join_args)
        if len(join_args) == 1:
            return join_args[0]
        elif len(join_args) > 0:
            L.debug("Joining %s args on %s", len(join_args), goal)
            join_args = sorted(join_args, key=len)
            res = join_args[0]
            res.intersection_update(*join_args[1:])
            L.debug("Joined %s(sizes=%s) args on %s. Result size = %s", len(join_args),
                [len(s) for s in join_args], goal, len(res))
            return res
        else:
            return EMPTY_SET

    def _qpr_helper(self, sub, search_triple, join_args):
        seen = set()
        try:
            L.debug("_qpr_helper: sub %s, search_triple %s", sub, search_triple)
            idx = search_triple.index(None)
            other_idx = 0 if (idx == 2) else 2
            qx = None

            if isinstance(search_triple[other_idx], Variable):
                sub_results = list(self.query_path_resolver(sub))
                L.debug("_qpr_helper: sub_results %s", sub_results)

                if idx == 2:
                    qx = (sub_results, search_triple[1], None)
                else:
                    qx = (None, search_triple[1], sub_results)

                if sub_results:
                    trips = self.triples_choices(qx)
                else:
                    trips = iter(())
            else:
                # join_args is assumed to be sorted such that it the most selective query was executed first, so we
                # should be able to profitably call triples_choices to reduce the size of our branch
                L.debug("_qpr_helper: joining...")
                if join_args:
                    # We use the last-added join_arg. It should be the smallest at this point
                    last_join = join_args[-1]
                    if last_join:
                        tl = (list(last_join),)
                        if idx == 2:
                            qx = search_triple[:2] + tl
                        else:
                            qx = tl + search_triple[1:3]
                        trips = self.triples_choices(qx)
                    else: # triples_choices treats [] as wildcard, but for us it's a 'match nothing', so...
                        trips = iter(())
                else:
                    qx = search_triple[:3]
                    trips = self.triples(qx)
            seen = set(y[idx] for y in trips)
            if L.isEnabledFor(logging.DEBUG):
                L.debug("_qpr_helper: Done with search_triple %s -> qx %s with %d seen", search_triple, qx, len(seen))
        finally:
            join_args.append(seen)

    def score(self, hop):
        if self.hop_scorer is not None:
            return self.hop_scorer(hop)
        return 0

    def triples_choices(self, query_triple):
        return self.graph.triples_choices(query_triple)

    def triples(self, query_triple):
        return self.graph.triples(query_triple)

    def __call__(self):
        return self.do_query()


def _format_paths(paths):
    sio = six.StringIO()
    for path in paths:
        for triple in path:
            idx = triple.index(None)
            other_idx = 0 if (idx == 2) else 2
            direction = '' if idx == 2 else '^'
            print('->' + str(triple[1]) + direction + '->' + str(triple[other_idx]), file=sio, end='')
        print(file=sio)
    return sio.getvalue()


class ComponentTripler(object):

    """ Gets a set of triples that are connected to the given object by
    objects which have an identifier.

    The ComponentTripler does not query against a backing graph, but instead
    uses the properties attached to the object.
    """

    def __init__(self, start, traverse_undefined=False, generator=False):
        self.start = start
        self.seen = set()
        self.generator = generator
        self.traverse_undefined = traverse_undefined

    def g(self, current_node, i=0):
        if not self.see_node(current_node):
            if self.traverse_undefined or current_node.defined:
                for x in chain(self.recurse_upwards(current_node, i),
                               self.recurse_downwards(current_node, i)):
                    yield x

    def recurse_upwards(self, current_node, depth):
        for prop in current_node.owner_properties:
            for x in self.recurse(prop.owner, prop, current_node, UP, depth):
                yield x

    def recurse_downwards(self, current_node, depth):
        for prop in current_node.properties:
            for val in prop.values:
                for x in self.recurse(current_node, prop, val, DOWN, depth):
                    yield x

    def recurse(self, lhs, via, rhs, direction, depth):
        (ths, nxt) = (rhs, lhs) if direction is UP else (lhs, rhs)
        if self.traverse_undefined or nxt.defined:
            yield (lhs.idl, via.link, rhs.idl)
            for x in self.g(nxt, depth + 1):
                yield x

    def see_node(self, node):
        node_id = id(node)
        if node_id in self.seen:
            return True
        else:
            self.seen.add(node_id)
            return False

    def __call__(self):
        x = self.g(self.start)
        if self.generator:
            return x
        else:
            return set(x)


class _QueryPathElement(tuple):

    def __new__(cls):
        return tuple.__new__(cls, ([], []))

    @property
    def subpaths(self):
        return self[0]

    @subpaths.setter
    def subpaths(self, toset):
        del self[0][:]
        self[0].extend(toset)

    @property
    def path(self):
        return self[1]


class _QueryPreparer(object):

    def __init__(self, start):
        self.seen = list()
        self.stack = list()
        self.paths = list()
        self.start = start
        self.variables = dict()
        self.vcount = 0
        # TODO: Refactor. The return values are not actually
        # used for anything

    def gather_paths_along_properties(
            self,
            current_node,
            property_list,
            direction):
        L.debug("gpap: current_node %s", current_node)
        ret = []
        is_good = False
        for this_property in property_list:
            L.debug("this_property is %s", this_property)
            if direction is UP:
                others = [this_property.owner]
            else:
                others = this_property.values

            for other in others:
                other_id = other.idl

                if isinstance(other, InRange):
                    other_id = _Range(other.min_value, other.max_value)
                elif not other.defined:
                    other_id = self.var(other_id)

                if direction is UP:
                    self.stack.append((other_id, this_property.link, None,
                                      current_node))
                else:
                    self.stack.append((None, this_property.link, other_id,
                                      current_node))
                L.debug("gpap: preparing %s from %s", other, this_property)
                subpath = self.prepare(other)

                if len(self.stack) > 0:
                    self.stack.pop()

                if subpath[0]:
                    is_good = True
                    subpath[1].path.insert(
                        0, (current_node.idl, this_property, other.idl))
                    ret.insert(0, subpath[1])

        L.debug("gpap: exiting %s", "good" if is_good else "bad")
        return is_good, ret

    def var(self, v):
        if v in self.variables:
            return self.variables[v]
        else:
            var = Variable(self.vcount)
            self.variables[v] = var
            self.vcount += 1
            return var

    def prepare(self, current_node):
        L.debug("prepare: current_node %s", repr(current_node))
        if current_node.defined or isinstance(current_node, InRange):
            if len(self.stack) > 0:
                self.paths.append(list(self.stack))
            return True, _QueryPathElement()
        else:
            if current_node in self.seen:
                return False, _QueryPathElement()
            else:
                self.seen.append(current_node)
            owner_parts = self.gather_paths_along_properties(
                current_node,
                current_node.owner_properties,
                UP)
            owned_parts = self.gather_paths_along_properties(
                current_node,
                current_node.properties,
                DOWN)

            self.seen.pop()
            subpaths = owner_parts[1] + owner_parts[1]
            if len(subpaths) == 1:
                ret = subpaths[0]
            else:
                ret = _QueryPathElement()
                ret.subpaths = subpaths
            return (owner_parts[0] or owned_parts[0], ret)

    def __call__(self):
        x = self.prepare(self.start)
        L.debug("self.prepare() result:" + str(x))
        L.debug("_QueryPreparer paths:" + str(_format_paths(self.paths)))
        return self.paths


class DescendantTripler(object):

    """ Gets triples that the object points to, optionally transitively. """

    def __init__(self, start, graph=None, transitive=True):
        """
        Parameters
        ----------
        start : GraphObject
            the node to start from
        graph : rdflib.graph.Graph, optional
            if given, the graph to draw descedants from. Otherwise the object
            graph is used
        """
        self.seen = set()
        self.seen_edges = set()
        self.start = start
        self.graph = graph
        self.results = list()
        self.transitve = transitive

    def g(self, current_node):
        if current_node in self.seen:
            return
        else:
            self.seen.add(current_node)

        if not current_node.defined:
            return

        if self.graph is not None:
            for triple in self.graph.triples((current_node.idl, None, None)):
                self.results.append(triple)
                if self.transitve:
                    self.g(_DTWrapper(triple[2]))
        else:
            for e in current_node.properties:
                if id(e) not in self.seen_edges:
                    self.seen_edges.add(id(e))
                    for val in e.values:
                        if val.defined:
                            self.results.append((current_node.idl, e.link, val.idl))
                            if self.transitve:
                                self.g(val)

    def __call__(self):
        self.g(self.start)
        return self.results


class _DTWrapper():
    """ Used by DescendantTripler to wrap identifiers in GraphObjects """
    defined = True
    __slots__ = ['idl']

    def __init__(self, ident):
        self.idl = ident

    def __hash__(self):
        return hash(self.idl)

    def __eq__(self, other):
        if type(other) == type(self):
            return (other is self) or (other.idl == self.idl)
        else:
            return False


class LegendFinder(object):

    """ Gets a list of the objects which can not be deleted freely from the
    transitive closure.

    Essentially, this is the 'mark' phase of the "mark-and-sweep" garbage
    collection algorithm.

    "Heroes get remembered, but legends never die."
    """

    def __init__(self, start, graph=None):
        self.talked_about = dict()
        self.seen = set()
        self.start = start
        self.graph = graph

    def legends(self, o, depth=0):
        if o in self.seen:
            return
        self.seen.add(o)
        for prop in o.properties:
            for value in prop.values:
                if value != self.start:
                    count = self.count(value)
                    self.talked_about[value] = count - 1
                    self.legends(value, depth + 1)

    def count(self, o):
        if o in self.talked_about:
            return self.talked_about[o]
        else:
            i = 0
            if self.graph is not None:
                for _ in self.graph.triples((None, None, o.idl)):
                    i += 1
            else:
                for prop in o.owner_properties:
                    if prop.owner.defined:
                        i += 1
            return i

    def __call__(self):
        self.legends(self.start)
        return {x for x in self.talked_about if self.talked_about[x] > 0}


class HeroTripler(object):

    def __init__(self, start, graph=None, legends=None):
        self.seen = set()
        self.start = start
        self.heroslist = set()
        self.results = set()
        self.graph = graph

        if legends is None:
            self.legends = LegendFinder(self.start, graph)()
        else:
            self.legends = legends

    def isLegend(self, o):
        return o in self.legends

    def isHero(self, o):
        return o in self.heroslist

    def heros(self, o, depth=0):
        if o in self.seen:
            return
        self.seen.add(o)

        for prop in o.properties:
            for value in prop.values:
                if not self.isLegend(value):
                    self.heros(value, depth + 1)
                    self.hero(value)

    def hero(self, o):
        if not self.isHero(o):
            if self.graph is not None:
                for trip in self.graph.triples((o.idl, None, None)):
                    self.results.add(trip)
            else:
                for e in o.properties:
                    for val in e.values:
                        if val.defined:
                            self.results.add((o.idl, e.link, val.idl))
            self.heroslist.add(o)

    def __call__(self):
        self.heros(self.start)
        self.hero(self.start)
        return self.results


class ReferenceTripler(object):

    def __init__(self, start, graph=None):
        self.seen = set()
        self.seen_edges = set()
        self.start = start
        self.results = set()
        self.graph = graph

    def refs(self, o):
        if self.graph is not None:
            for trip in chain(
                self.graph.triples(
                    (None, None, o.idl)),
                self.graph.triples(
                    (o.idl, None, None))):
                self.results.add(trip)
        else:
            for e in o.properties:
                if (DOWN, id(e)) not in self.seen_edges:
                    self.seen_edges.add((DOWN, id(e)))
                    for val in e.values:
                        if val.defined:
                            self.results.add((o.idl, e.link, val.idl))

            for e in o.owner_properties:
                if (UP, id(e)) not in self.seen_edges:
                    self.seen_edges.add((UP, id(e)))
                    if e.owner.defined:
                        self.results.add((e.owner.idl, e.link, o.idl))

    def __call__(self):
        self.refs(self.start)
        return self.results


class IdentifierMissingException(Exception):

    """ Indicates that an identifier should be available for the object in
        question, but there is none """

    def __init__(self, dataObject="[unspecified object]", *args, **kwargs):
        super(IdentifierMissingException, self).__init__(
            "An identifier should be provided for {}".format(str(dataObject)),
            *args,
            **kwargs)



import logging

import rdflib

from .graph_object import GraphObjectQuerier
from .rdf_query_modifiers import (ZeroOrMoreTQLayer,
                                  rdfs_subclassof_zom_creator)

L = logging.getLogger(__name__)


def goq_hop_scorer(hop):
    if hop[1] == rdflib.RDF.type:
        return 1
    return 0


def load_base(graph, idents, target_type, context, resolver):
    '''
    Loads a set of objects from an RDF graph given their identifiers

    Parameters
    ----------
    graph : rdflib.graph.Graph
        The graph to query from
    idents : list of rdflib.term.URIRef
        A list of identifiers to convert into objects
    target_type : rdflib.term.URIRef
        URI of the target type. Any result will be a sub-class of this type
    context : object
        Limits the scope of the query to statements within or entailed by this context.
        Notionally, it's a owmeta_core.context.Context instance
    resolver : .rdf_type_resolver.RDFTypeResolver
        Handles some of the mappings
    '''

    L.debug("load_base: graph %s target_type %s context %s resolver %s",
            graph, target_type, context, resolver)
    if not idents:
        return

    grouped_types = dict()
    # We don't use a subclassof ZOM layer for this query since we are going to get the
    # "most specific" type, which will have to be one declared explicitly
    L.debug("querying %s types in %s", idents, graph)
    idents = list(idents)
    ids_missing_types = set(idents)
    for ident, _, rdf_type in graph.triples_choices((idents,
                                                     rdflib.RDF['type'],
                                                     None)):
        t = grouped_types.get(ident, None)
        if t is None:
            ids_missing_types.remove(ident)
            grouped_types[ident] = set([rdf_type])
        else:
            t.add(rdf_type)
    if ids_missing_types:
        L.debug('Could not recover type declarations for %r. Defaulting to %s',
                ids_missing_types, rdflib.RDFS.Resource)
        for ident in ids_missing_types:
            grouped_types[ident] = set([rdflib.RDFS.Resource])

    for ident, types in grouped_types.items():
        the_type = resolver.type_resolver(graph, types, base=target_type)
        if the_type is None:
            raise MissingRDFTypeException(
                    f'The type resolver could not recover an RDF type for {ident}'
                    f' from {types} in {graph} constrained to sub-classes of {target_type}')
        yield resolver.id2ob(ident, the_type, context)


def load_terms(graph, start, target_type):
    '''
    Loads a set of terms based on the object graph starting from `start`

    Parameters
    ----------
    graph : rdflib.graph.Graph
        The graph to query from
    start : .graph_object.GraphObject
        The graph object to start the query from
    target_type : rdflib.term.URIRef
        URI of the target type. Any result will be a sub-class of this type
    '''

    L.debug("load: start %s, target_type %s, graph %s", start, target_type, graph.store)
    graph = ZeroOrMoreTQLayer(rdfs_subclassof_zom_creator(target_type), graph)
    return GraphObjectQuerier(start, graph, hop_scorer=goq_hop_scorer)()


def load(graph, start, target_type, *args):
    '''
    Loads a set of objects based on the graph starting from `start`

    Parameters
    ----------
    graph : rdflib.graph.Graph
        The graph to query from
    start : .graph_object.GraphObject
        The graph object to start the query from
    target_type : rdflib.term.URIRef
        URI of the target type. Any result will be a sub-class of this type
    '''

    idents = load_terms(graph, start, target_type)
    return load_base(graph, idents, target_type, *args)


def get_most_specific_rdf_type(graph, types, base=None):
    '''
    Find the RDF type that isn't a sub-class of any other, constrained to be a sub-class
    of `base` if that is provided.

    Parameters
    ----------
    graph : rdflib.graph.Graph
        The graph to query rdfs:subClassOf relationships
    types : list of rdflib.term.URIRef
        The types to query
    base : rdflib.term.URIRef
        The "base" type

    See Also
    --------
    RDFTypeResolver
    '''
    if len(types) == 1 and (not base or (base,) == tuple(types)):
        return tuple(types)[0]

    if not types and base:
        return base

    most_specific_types = _gmsrt_helper(graph, types, base)

    if len(most_specific_types) == 1:
        return most_specific_types.pop()
    else:
        L.warning(('No most-specific type could be determined among %s'
                   ' constrained to subclasses of %r'), types, base)
        return None


def _gmsrt_helper(graph, start, base=None):
    # Finds the most specific (furthest to the left in the chain of subClassOf
    # relationships), then confirms that the resulting set is a sub-class of the base if
    # one is given
    res = set(start)
    border = set(start)
    subclasses = {s: {s} for s in start}
    hit = True
    while len(res) > 1:
        new_border = set()
        itr = graph.triples_choices((list(border), rdflib.RDFS.subClassOf, None))
        hit = False
        for item in itr:
            if isinstance(item[0], tuple):
                # If we retrieved from a rdflib.store.Store instead of a Graph, then we
                # have to get the first element of the pair `item` to get the actual triple
                triple = item[0]
            else:
                triple = item
            subj, _, obj = triple
            if obj != subj:
                obj_subclasses = subclasses.get(obj, None)
                if obj_subclasses is None:
                    subclasses[obj] = obj_subclasses = set()
                obj_subclasses |= subclasses[subj]
                res.discard(obj)
                hit = True
                new_border.add(obj)
        if not hit:
            break
        border = new_border
    if base is not None:
        if not hit:
            # If hit is False, then we've stopped because no more super-classes were found, so
            # base *has* to be in here or else none of our types are any good
            res &= subclasses.get(base, set())
        else:
            # We exited because we eliminated all of the other types (or only had one to
            # begin with), but keep going to make sure we have the base.
            seen = set(border)
            while border:
                if base in border:
                    break
                border = {o for _, _, o
                        in graph.triples_choices((list(border), rdflib.RDFS.subClassOf, None))
                        if o not in seen}
                seen |= border
            else: # no break
                # We never found the base class in super-classes of our result set, so we
                # have to drop the result set
                res = set()
    return res


def oid(identifier_or_rdf_type, rdf_type, context, base_type=None):
    """
    Create an object from its rdf type

    Parameters
    ----------
    identifier_or_rdf_type : rdflib.term.URIRef
        If `rdf_type` is provided, then this value is used as the identifier
        for the newly created object. Otherwise, this value will be the
        :attr:`rdf_type` of the object used to determine the Python type and
        the object's identifier will be randomly generated.
    rdf_type : rdflib.term.URIRef
        If provided, this will be the :attr:`rdf_type` of the newly created
        object.
    context : Context, optional
        The context to resolve a class from
    base_type : type
        The base type

    Returns
    -------
       The newly created object

    """
    identifier = identifier_or_rdf_type
    if rdf_type is None:
        rdf_type = identifier_or_rdf_type
        identifier = None

    cls = None
    if context is not None:
        cls = context.resolve_class(rdf_type)

    if cls is None and context is not None:
        for types in _superclass_iter(context.rdf_graph(), rdf_type):
            for typ in types:
                cls = context.resolve_class(typ)
                if cls is not None:
                    break
            if cls is not None:
                break

    if cls is None:
        cls = base_type

    # if its our class name, then make our own object
    # if there's a part after that, that's the property name
    if context is not None:
        cls = context(cls)

    if identifier is not None:
        o = cls.query(ident=identifier, no_type_decl=True)
    else:
        o = cls.query()
    return o


def _superclass_iter(graph, start):
    '''
    Generate up the super-classes of this type
    '''
    border = set([start])
    seen = set([start])
    while True:
        new_border = set()

        for t in graph.triples_choices((list(border), rdflib.RDFS.subClassOf, None)):
            if t[2] not in seen:
                new_border.add(t[2])
                seen.add(t[2])

        if border == new_border:
            break
        if not new_border:
            break
        yield new_border
        border = new_border


class MissingRDFTypeException(Exception):
    '''
    Raised when we were looking for an RDF type couldn't find one
    '''
